
local _,rematch = ...

--[[ Floating Pet Card ]]

-- desturates floating pet cards that are not owned
local function SetCardDesaturate(desat)
	local card = RematchFloatingPetCard
	card.leftAbilitiesBG:SetDesaturated(desat)
	card.rightAbilitiesBG:SetDesaturated(desat)
	card.leftTitleBG:SetDesaturated(desat)
	card.rightTitleBG:SetDesaturated(desat)
	card.lineAbilitiesBG:SetDesaturated(desat)
	card.lineTitleBG:SetDesaturated(desat)
	if desat then
		card.back:SetBackdropBorderColor(.75,.75,.75)
		card:SetBackdropBorderColor(.75,.75,.75)
	else
		card.back:SetBackdropBorderColor(1,.82,0)
		card:SetBackdropBorderColor(1,.82,0)
	end
end

rematch.vulnerabilities = {{4,5},{1,3},{6,8},{5,2},{8,7},{2,9},{9,10},{10,1},{3,4},{7,6}}

function rematch:ShowFloatingPetCard(petID,anchorPoint,relativeTo,relativePoint,xoff,yoff)

	local card = RematchFloatingPetCard
	card:SetPoint(anchorPoint,relativeTo,relativePoint,xoff,yoff)

	card.back:SetShown(IsAltKeyDown())

	if petID and petID==card.petID then
		card:Show() -- we've done all this already, just show the card
		return
	end
	card.petID = petID

	card:SetParent(relativeTo)

	local speciesID,icon,customName,realName,level,xp,maxXp,displayID,isFavorite,icon,petType,sourceText,description,canBattle,creatureID

	if type(petID)=="string" then
		speciesID,customName,level,xp,maxXp,displayID,isFavorite,realName,icon,petType,creatureID,sourceText,description,_,canBattle = C_PetJournal.GetPetInfoByPetID(petID)
	elseif type(petID)=="number" then
		speciesID = petID
		petID = nil
		realName,icon,petType,creatureID,sourceText,description,_,canBattle = C_PetJournal.GetPetInfoBySpeciesID(speciesID)
	else
		return
	end

	card.back.source:SetText(sourceText)
	card.back.description:SetText(description)

	-- update model
	local model = card.model
	if displayID and displayID~=0 then
		if ( displayID ~= model.displayID ) then
			model.creatureID = nil
			model.displayID = displayID
			model:SetDisplayInfo(displayID)
			model:SetDoBlend(false)
		end
	elseif creatureID and creatureID~=0 then
		if ( creatureID ~= model.creatureID ) then
			model.creatureID = creatureID;
			model.displayID = nil;
			model:SetCreature(creatureID);
			model:SetDoBlend(false);
		end
	end
	model:SetAnimation(0)

--	card.icon:SetNormalTexture(icon)
	SetPortraitToTexture(card.icon.icon,icon)
	card.petType.icon:SetTexture("Interface\\PetBattles\\PetIcon-"..PET_TYPE_SUFFIX[petType])
	card.name:SetText(customName or realName)

	if customName then
		card.realName:SetText(realName)
		card.realName:Show()
	else
		card.realName:Hide()
	end

	if speciesID and canBattle then
--		print(self.speciesID,"can battle")
		C_PetJournal.GetPetAbilityList(speciesID,rematch.abilityList,rematch.levelList)
		for i=1,#rematch.abilityList do
			local _,name, icon, _, _, _, petType, noHints = C_PetBattles.GetAbilityInfoByID(rematch.abilityList[i])
			local index = tostring(i)
			card.abilities[index].type:SetTexture("Interface\\PetBattles\\PetIcon-"..PET_TYPE_SUFFIX[petType])
			card.abilities[index].name:SetText(name)
		end
		card.cantbattle:Hide()
		card.abilities:Show()
		if petID then -- this is a pet we own with actual stats

			SetCardDesaturate(false)
			card.stats.level:SetText(level)

			local health, maxHealth, power, speed, rarity = C_PetJournal.GetPetStats(petID)
			if health==0 then
				card.stats.hpBar:Hide()
				card.stats.health:SetText("Dead")
				model:SetAnimation(1) -- death animation
			else
				card.stats.hpBar:SetWidth(48*(health/maxHealth))
				card.stats.hpBar:Show()
				card.stats.health:SetText(maxHealth)
			end
			card.stats.power:SetText(power)
			card.stats.speed:SetText(speed)

			local color = ITEM_QUALITY_COLORS[rarity-1]
			card.stats.quality:SetText(_G["BATTLE_PET_BREED_QUALITY"..rarity])
			card.stats.quality:SetVertexColor(color.r,color.g,color.b)

			local isBreedable = IsAddOnLoaded("BattlePetBreedID")
			card.stats.breed:SetText(isBreedable and GetBreedID_Journal(petID) or "")
			card.stats.breed:SetShown(isBreedable)
			card.stats.breedIcon:SetShown(isBreedable)

			if level<25 then
				card.stats.xpBar:SetMinMaxValues(0,maxXp)
				card.stats.xpBar:SetValue(xp)
				card.stats.xpBar.rankText:SetFormattedText(PET_BATTLE_CURRENT_XP_FORMAT_BOTH, xp, maxXp, xp/maxXp*100);
				card.stats.xpBar:Show()
				card.stats.pennant:SetVertexColor(1,1,1)
			else
				card.stats.xpBar:Hide()
				card.stats.pennant:SetVertexColor(1,.82,0)
			end

			card.stats:Show()
		else -- this is not a pet we own, hide stats
			SetCardDesaturate(true)
			card.stats:Hide()
		end
		-- set vulnerabilities
		card.back.strongType:SetTexture("Interface\\PetBattles\\PetIcon-"..PET_TYPE_SUFFIX[rematch.vulnerabilities[petType][1]])
		card.back.weakType:SetTexture("Interface\\PetBattles\\PetIcon-"..PET_TYPE_SUFFIX[rematch.vulnerabilities[petType][2]])

	else -- this pet can't battle
		if petID then
			SetCardDesaturate(false)
		else
			SetCardDesaturate(true)
		end
		card.abilities:Hide()
		card.stats:Hide()
		card.cantbattle:Show()
	end
	card:Show()
end

function rematch:HideFloatingPetCard()
	RematchFloatingPetCard:Hide()
end
