
local _,rematch = ...
Rematch = rematch -- for xml's use, addon namespace
local frame = RematchFrame -- for lua's use, primary frame defined in xml

RematchSaved = {} -- saved teams SavedVariable
-- teams are indexed by name, either the npc name or a user-chosen name
-- ["Name"] = { {pet1,a1,a2,a3},{pet2,a1,a2,a3},{pet3,a1,a2,a3} }
-- if an npcID is known:
-- ["Name"] = { {pet1,a1,a2,a3},{pet2,a1,a2,a3},{pet3,a1,a2,a3},npcID }
-- Pets can be species IDs (numeric), usually for pets received or imported.
-- When sending pets, they need to be sent as species IDs
-- They're resolved into petIDs when loaded

RematchSettings = {} -- savedvariables settings

local saved -- will be RematchSaved at PLAYER_LOGIN
local settings -- will be RematchSettings at PLAYER_LOGIN

local selectedTeam -- team selected in drawer list
local targetedNpcID -- npcID of targeted unit
local targetedName -- name of targeted unit

rematch.showDrawer = nil -- whether team/options are displayed in a longer frame
rematch.shareMode = nil -- 1=export, 2=import, 3=send

rematch.teamList = {} -- numerically-indexed list of teams (just name or npcID, index to saved) for the list
rematch.abilityList = {} -- reusable table for GetPetAbilityList
rematch.levelList = {} -- reusable table for GetPetABilityList
rematch.info = {} -- reusable generic table (to reduce table/garbage creation)
rematch.current = {} -- current pets
rematch.attackTypes = {} -- ability types that can attack, indexed by type, for generating hints
rematch.petsToLoad = {{},{},{}} -- used in loading process
rematch.incomingTeam = {}

-- attack types (index) are {strong vs type,weak vs type}
-- ie flying(3rd) is strong vs aquatic(9), weak vs dragonkin(2)
rematch.attackHints = { {2,8},{6,4},{9,2},{1,9},{4,1},{3,10},{10,5},{5,3},{7,6},{8,7} }

rematch.optionsText = {
	KeepFirst={"Keep first pet","For leveling pets, only load the first two saved pets alongside the current first pet. This can also be set by clicking the lock on the first pet."},
	AutoShow={"Auto show","When targeting something with saved pets, and the pets are not already loaded, show the Rematch window and close it when pets load."},
	AutoLoad={"Auto load","When you target something with saved pets, and the pets are not already loaded, load the pets immediately."},
	KeepSummoned={"Keep companion","After a team is loaded, summon back the companion that was at your side before the load."},
	DisableShare={"Disable sharing","Disable the Send button and also block any incoming pets sent by others. Import and Export still work."},
	LargeWindow={"Larger window","Make the Rematch window larger for easier viewing."},
	LockWindow={"Lock window","Lock the Rematch window so it can't be moved or dismissed with the Escape key."},
}

rematch.shareText = {
	{"Exporting","The text below describes this team in a format usable by other users. Hit Ctrl+C to copy it to the clipboard. Ctrl+V to paste it into an email, forum or elsewhere."},
	{"Importing","If you have a team copied to the clipboard, paste it here with Ctrl+V. You can review the team above before hitting Enter or clicking Save to save the team."},
	{"Sending","Who would you like to send this team to?"},
}

BINDING_HEADER_REMATCH = "Rematch"

function rematch:OnEvent(event,...)
	if rematch[event] then
		rematch[event](rematch,...)
	end
end
frame:RegisterEvent("PLAYER_LOGIN")

function rematch:PLAYER_LOGIN()
	saved = RematchSaved
	settings = RematchSettings
	if saved.Settings then -- this has old rematch savedvariables, upgrade it
		self:UpgradeSavedVars()
	end
	-- if savedvar exists, position window
	if settings.XPos then
		frame:SetScale(settings.LargeWindow and 1.25 or 1)
		frame:ClearAllPoints()
		frame:SetPoint("BOTTOMLEFT",settings.XPos,settings.YPos)
	end
	self:BuildUI()
	self:OrderTeamList()
	-- setup HybridScrollFrame
	local scrollFrame = frame.drawer.teams.scrollFrame
	scrollFrame.update = self.UpdateTeamList
	scrollFrame.stepSize = 54
	HybridScrollFrame_CreateButtons(scrollFrame,"RematchTeamTemplate")
	frame:RegisterEvent("PLAYER_TARGET_CHANGED")
	frame:RegisterEvent("PLAYER_REGEN_DISABLED") -- to close rematch when entering combat
	frame:RegisterEvent("PLAYER_REGEN_ENABLED") -- to resume target watching when leaving combat
	RegisterAddonMessagePrefix("Rematch")
	frame:RegisterEvent("CHAT_MSG_ADDON")
end

function rematch:PLAYER_TARGET_CHANGED()
	targetedName,targetedNpcID = rematch:GetUnitNameandID("target")
	selectedTeam = nil -- ordinarily, clear selectedTeam when changing targets
	if saved[targetedName] and (not saved[targetedName][4] or saved[targetedName][4]==targetedNpcID) then
		-- but if we have an npc targeted with a matching name, set selectedTeam to that target
		-- note this can set selectedTeam to named players with matching team names (which is ok)
		selectedTeam = targetedName
	end

	if selectedTeam and (settings.AutoShow or settings.AutoLoad) then
		local petsNeedLoading = rematch:PetsNeedLoading(selectedTeam)
		if petsNeedLoading then
			if settings.AutoShow then
				frame:Show()
				selectedTeam = targetedName
			end
			if settings.AutoLoad then
				rematch:LoadPets(selectedTeam)
			end
		end
	end

	if frame:IsVisible() then
		if frame.drawer:IsVisible() then
			rematch:UpdateTeamList()
		end
		rematch:UpdateSavedPets()
	end
end

function rematch:PLAYER_REGEN_DISABLED()
	frame:UnregisterEvent("PLAYER_TARGET_CHANGED")
	if frame:IsVisible() then
		frame:Hide()
	end
end

function rematch:PLAYER_REGEN_ENABLED()
	frame:RegisterEvent("PLAYER_TARGET_CHANGED")
end

-- these events are only registered while the rematch window is up

-- fired when pets are healed, resurrected, etc
function rematch:UPDATE_SUMMONPETS_ACTION()
	rematch:UpdateCurrentPets()
end

-- track when Revive Battle Pets & Battle Pet Bandage is used
function rematch:UNIT_SPELLCAST_SUCCEEDED(unit,_,_,_,spellID)
	-- 125439 is revive pet spell
	if unit=="player" and spellID==125439 then -- revive battle pet
		Rematch:StartTimer("UpdateHealButtons",0.1,rematch.UpdateHealButtons)
	end
end

-- can't just track when bandages used, also need to track when bandages received
-- this is only registered while the rematch window is visible
function rematch:BAG_UPDATE()
	rematch:UpdateHealButtons()
end

function rematch:UpdateHealButtons()
	frame.petheal.cooldown:SetCooldown(GetSpellCooldown(125439))
	local count = GetItemCount(86143)
	if count==0 then
		frame.petbandage.icon:SetVertexColor(.5,.5,.5)
	else
		frame.petbandage.icon:SetVertexColor(1,1,1)
	end
	frame.petbandage.count:SetText(GetItemCount(86143))
end

function rematch:HealButtonOnEnter()
	GameTooltip:SetOwner(self,"ANCHOR_RIGHT")
	if self:GetAttribute("type")=="spell" then
		GameTooltip:SetSpellByID(125439)
	else
		GameTooltip:SetItemByID(86143)
	end
	GameTooltip:Show()
end

-- Upgrade Rematch 1.x savedvar for 2.x: RematchSaved for teams, RematchSettings for settings
function rematch:UpgradeSavedVars()
	for k,v in pairs(saved) do
		if k=="Settings" then
			for x,y in pairs(v) do
				settings[x]=y
			end
		end
	end
	saved.Settings = nil
end

function frame:Toggle()
	if InCombatLockdown() then
		print("\124cffffff00You're in combat. Blizzard has restrictions on what we can do with pets during combat. Try again when you're out of combat. Sorry!")
	else
		frame:SetShown(not frame:IsVisible())
	end
end

function rematch:BuildUI()
	frame.savedHeader.text:SetText("Saved for <no target>")
	frame.saveAs.header.text:SetText("Save Team As...")
	frame.options.header.text:SetText("Options")
	frame.options.version:SetText("Rematch version "..GetAddOnMetadata("Rematch","Version"))
	local modalLevel = frame.drawer.teams:GetFrameLevel()+4
	frame.saveAs:SetFrameLevel(modalLevel)
	frame.options:SetFrameLevel(modalLevel)
	frame.share:SetFrameLevel(modalLevel)

	-- current pet slots+health bars and their abilities
	frame.current = {} -- 3 current pet slots
	for i=1,3 do
		frame.current[i] = CreateFrame("Button",nil,frame,"RematchLargeSlotTemplate")
		local button = frame.current[i]
		button:SetPoint("TOPLEFT",(i-1)*78+10,-20)
		frame.current[i].ability = {} -- 3 abilities of each current pet
		for j=1,3 do
			button.ability[j] = CreateFrame("Button",nil,frame.current[i],"RematchSmallSlotTemplate")
			button.ability[j]:SetPoint("TOPLEFT",frame.current[i],"TOPRIGHT",2,(j-1)*-22)
		end
--		button.health:SetVertexColor(0,.85,0,1)
	end

	-- saved pet slots
	frame.saved = {}
	for i=1,3 do
		frame.saved[i] = CreateFrame("Button",nil,frame,"RematchMediumSlotTemplate")
		local button = frame.saved[i]
		button:SetPoint("TOPLEFT",(i-1)*66+10,-121)
		frame.saved[i].ability = {}
		for j=1,3 do
			button.ability[j] = CreateFrame("Button",nil,frame.saved[i],"RematchTinySlotTemplate")
			button.ability[j]:SetPoint("TOPLEFT",frame.saved[i],"TOPRIGHT",2,(j-1)*-16+5)
		end
	end

	-- add lock on first pet
	frame.current[1].lock = CreateFrame("CheckButton",nil,frame.current[1],"RematchPetLock")
	frame.current[1].lock:SetPoint("BOTTOMLEFT",-8,-4)

	-- assign icons for Revive Battle Pets (125439) and Battle Pet Bandage (86143)
	frame.petheal.icon:SetTexture((GetSpellTexture(125439)))
	frame.petbandage.icon:SetTexture(GetItemIcon(86143))

	-- go through each checkbutton setting, set its text and run its reflect
	for setting,opt in pairs(rematch.optionsText) do
		frame.options.checks[setting].text:SetText(opt[1])
		frame.options.checks[setting].tooltip = opt[2]
		frame.options.checks[setting].setting = setting
		if rematch.ReflectSetting[setting] then
			rematch.ReflectSetting[setting]()
		end
	end

end

function rematch:PetLockOnEnter()
-- **
--	if not RematchFloatingPetCard:IsVisible() then
--		rematch:ShowFloatingPetCard(self:GetParent().petID,"BOTTOMLEFT",frame.current[1].ability[3],"BOTTOMRIGHT",0,0)
--	end
end

function rematch:PetLockOnLeave()
	if not settings.KeepFirst then
		frame.current[1].lock:Hide()
	end
	if not MouseIsOver(frame.current[1]) then
		rematch:HideFloatingPetCard()
	end
end

function rematch:PetLockOnClick()
	settings.KeepFirst = self:GetChecked()
	rematch.ReflectSetting.KeepFirst()
end

function rematch:PetOnEnter()
	if self==frame.current[1] then
		frame.current[1].lock:Show()
	end
	rematch:ShowFloatingPetCard(self.petID,"BOTTOMLEFT",self.ability[3],"BOTTOMRIGHT",0,0)
end

function rematch:PetOnLeave()
	if self==frame.current[1] and not settings.KeepFirst and not MouseIsOver(frame.current[1].lock) then
		frame.current[1].lock:Hide()
	end
	if not MouseIsOver(frame.current[1].lock) then
		rematch:HideFloatingPetCard()
	end
end

function rematch:PetOnClick()
	if self.petID then
		-- shift+click to link pet (only known pets work--default can't shift+click missing either)
		if IsModifiedClick("CHATLINK") and type(self.petID)=="string" then
			ChatEdit_InsertLink(C_PetJournal.GetBattlePetLink(self.petID))
			return
		end
		-- otherwise bring up pet journal and select the chosen pet/species
		if not PetJournalParent then
			PetJournal_LoadUI()
		end
		if not PetJournalParent:IsShown() then
			ShowUIPanel(PetJournalParent)
		elseif PetJournalPetCard.petID==self.petID or PetJournalPetCard.speciesID==self.petID then
			HideUIPanel(PetJournalParent)
		end
		PetJournalParent_SetTab(PetJournalParent,2)
		if type(self.petID)=="string" then
			PetJournal_SelectPet(PetJournal,self.petID)
		else -- petID is a species, a missing pet
			PetJournal_SelectSpecies(PetJournal,self.petID)
		end
	end
end

function rematch:AbilityOnClick()
	if self.abilityID then
		if IsModifiedClick("CHATLINK") then
			local petID = self:GetParent().petID
			local maxHealth,power,speed,_ = 100,0,0
			if type(petID)=="string" then
				_,maxHealth,power,speed = C_PetJournal.GetPetStats(petID)
			end
			ChatEdit_InsertLink(GetBattlePetAbilityHyperlink(self.abilityID,maxHealth,power,speed))
		elseif self:GetParent().petID then
			rematch.PetOnClick(self:GetParent()) -- pretend we clicked parent pet
		end
	end
end
function rematch:VerifyButtons()
	frame.saveButton:Enable()
	frame.deleteButton:Enable()
	frame.loadButton:Disable()
	frame.drawer.sendButton:Disable()
	frame.drawer.exportButton:Disable()
	if selectedTeam then
		if frame.drawer:IsVisible() then
			if not settings.DisableShare then
				frame.drawer.sendButton:Enable()
			end
			frame.drawer.exportButton:Enable()
			if not rematch.selectedTeamVisible then
				-- if team selected, drawer open and selected team not on list, disable for safety
				frame.deleteButton:Disable()
				frame.saveButton:Disable()
			end
		end
	elseif targetedNpcID and saved[targetedName] then
		-- do nothing, keep buttons enabled
	else
		frame.deleteButton:Disable()
	end
	if frame.saved[1].petID then
		frame.loadButton:Enable()
	end
end

-- panel buttons like close/load/save/delete
-- this is too spaghetti - each button has a name, break it out into a table like events?
function rematch:ButtonOnClick()
	if self==frame.closeButton then
		frame:Hide()
	elseif self==frame.saveButton then
		if not selectedTeam then
			if not targetedNpcID then -- save hit with no npc target (and no team selected)
				frame.drawer:Hide()
				frame.saveAs:Show()
			else -- save hit while npc targeted (and no team selected)
				rematch:SaveCurrentTeam(targetedName,targetedNpcID)
			end
		else -- hitting save with a team selected
			local npcID
			if not targetedName and saved[selectedTeam][4] then
				npcID = saved[selectedTeam][4] -- if nothing targeted, reuse existing npcID
			elseif targetedName==selectedTeam then
				npcID = targetedNpcID -- if something targeted, and it's same name, use targeted npcID
			end
			rematch:SaveCurrentTeam(selectedTeam,npcID)
		end
		if frame.drawer:IsVisible() then
			rematch:UpdateTeamList()
		end
	elseif self==frame.saveAs.cancelButton then
		frame.saveAs:Hide()
	elseif self==frame.saveAs.saveButton then
		local teamName = frame.saveAs.editBox:GetText()
		if saved[teamName] then
			frame.saveAs.warning:Show()
		else
			rematch:SaveCurrentTeam(teamName,nil)
			frame.saveAs:Hide()
		end
	elseif self==frame.deleteButton then
		if selectedTeam then
			saved[selectedTeam] = nil
			selectedTeam = nil
			rematch:UpdateSavedPets()
			rematch:OrderTeamList()
			rematch:UpdateTeamList()
		end
	elseif self==frame.loadButton then
		if selectedTeam and saved[selectedTeam] then
			rematch:LoadPets(selectedTeam)
		elseif targetedName and saved[targetedName] then
			rematch:LoadPets(targetedName)
		end
	elseif self==frame.drawer.optionsButton then
		frame.options:Show()
	elseif self==frame.options.okayButton then
		frame.options:Hide()
	elseif self==frame.drawer.exportButton then
		if selectedTeam then
			rematch:ExportTeam()
		end
	elseif self==frame.share.cancelButton then
		frame.share:Hide()
	elseif self==frame.drawer.importButton then
		rematch:ImportTeam()
	elseif self==frame.share.saveButton then
		rematch:SaveShare()
	elseif self==frame.drawer.sendButton then
		if selectedTeam then
			rematch.shareMode = 3
			frame.share:Show()
		end
	elseif self==RematchIncomingFrame.cancelButton then
		RematchIncomingFrame:Hide()
	elseif self==RematchIncomingFrame.saveButton then
		rematch:SaveIncoming()
	end
end

function rematch:SaveAsEditBoxChanged()
	if self:GetText():len()>0 then
		frame.saveAs.saveButton:Enable()
	else
		frame.saveAs.saveButton:Disable()
	end
	frame.saveAs.warning:Hide()
end

--[[ OnShow/OnHides ]]

-- esc is heavily used in rematch to close panels. when the drawer is shown,
-- it disables esc on parent frame and enables it on drawer. when options from drawer
-- is shown, it disables esc on drawer and enables it on options.
-- frame:Hide() doesn't run an OnHide handler if the frame is already hidden
-- remember that at any time the whole frame needs hidden due to entering combat

-- being in UISpecialFrames makes a frame hide on esc key
function rematch:SetESCable(frame,enable)
	if frame=="RematchFrame" and settings.LockWindow then
		enable = false -- disable esc for main window if LockWindow set
	end
	local hasFrame = tContains(UISpecialFrames,frame)
	if enable and not hasFrame then
		tinsert(UISpecialFrames,frame)
	elseif not enable and hasFrame then
		for i=1,#UISpecialFrames do
			if UISpecialFrames[i]==frame then
				table.remove(UISpecialFrames,i)
				return
			end
		end
	end
end

function rematch:OnShow()
	if not C_PetJournal.GetPetLoadOutInfo(1) then
		-- due to lag or from summoning the window before pets load, loadout pets may not exist.
		-- in those cases come back in half a second and try again
		rematch:StartTimer("PetsRanAway",.5,rematch.OnShow)
		return
	end
	rematch:PLAYER_TARGET_CHANGED() -- update target stuff
	rematch:UpdateCurrentPets()
	rematch:UpdateSavedPets()
	rematch:WatchForChangingPets()
	rematch:UpdateHealButtons()
	rematch:SetESCable("RematchFrame",true)
	frame:RegisterEvent("UPDATE_SUMMONPETS_ACTION")
	frame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	frame:RegisterEvent("BAG_UPDATE")
end

function rematch:OnHide()
	frame:UnregisterEvent("BAG_UPDATE")
	frame:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	frame:UnregisterEvent("UPDATE_SUMMONPETS_ACTION")
	frame.options:Hide()
	frame.share:Hide()
	frame.drawer:Hide()
	frame.saveAs:Hide()
	RematchFloatingPetCard:Hide()
	rematch:SetESCable("RematchFrame",false)
end

function rematch:DrawerOnShow()
	frame.saveAs:Hide()
	rematch.showDrawer = true
	rematch:OrderTeamList()
	rematch:UpdateTeamList()
	frame:SetHeight(342) -- was 332
	frame.toggle:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
	frame.toggle:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down")
	rematch:SetESCable("RematchFrame",false)
	rematch:SetESCable("RematchDrawer",true)
end

function rematch:DrawerOnHide()
	rematch.showDrawer = nil
	rematch:ClearSelected()
	settings.XPos,settings.YPos = frame:GetLeft(),frame:GetBottom()
	frame:SetHeight(196) -- was 188
	frame.toggle:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Up")
	frame.toggle:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Down")
	rematch:SetESCable("RematchDrawer",false)
	rematch:SetESCable("RematchFrame",true)
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMLEFT",settings.XPos,settings.YPos)
end

function rematch:SaveAsOnShow()
	rematch:SetESCable("RematchFrame",false)
	rematch:SetESCable("RematchSaveAs",true)
	frame.saveAs.saveButton:Disable()
	frame.savedHeader:Hide()
end

function rematch:SaveAsOnHide()
	rematch:SetESCable("RematchSaveAs",false)
	rematch:SetESCable("RematchFrame",true)
	frame.saveAs.editBox:SetText("")
	frame.savedHeader:Show()
end

function rematch:OptionsOnShow()
	rematch:SetESCable("RematchDrawer",false)
	rematch:SetESCable("RematchOptions",true)
	frame.savedHeader:Hide()
end

function rematch:OptionsOnHide()
	rematch:SetESCable("RematchOptions",false)
	rematch:SetESCable("RematchDrawer",true)
	frame.savedHeader:Show()
end

function rematch:ShareOnShow()
	rematch:SetESCable("RematchDrawer",false)
	rematch:SetESCable("RematchShare",true)
	frame.drawer.teams:Hide()
	local shareMode = rematch.shareMode
	frame.share.header.text:SetText(rematch.shareText[shareMode][1])
	frame.share.descBox.text:SetText(rematch.shareText[shareMode][2])
	frame.share.cancelButton:SetText(shareMode==1 and "Ok" or "Cancel")
	frame.share.detailBox.text:SetText("")
	frame.share.saveButton:SetText(shareMode==2 and "Save" or "Send")
	frame.share.saveButton:SetShown(shareMode~=1) -- don't show save/send button for Export
	rematch:SetShareStatus() -- clear any warning
end

function rematch:ShareOnHide()
	rematch:SetESCable("RematchShare",false)
	rematch:SetESCable("RematchDrawer",true)
	if selectedTeam and selectedTeam:match("^~temp~") then
		selectedTeam = nil
	end
	rematch:UpdateSavedPets()
	rematch:OrderTeamList()
	rematch:UpdateTeamList()
	frame.drawer.teams:Show()
	frame:UnregisterEvent("CHAT_MSG_SYSTEM") -- in case any sending was not finished
end

function rematch:SaveCurrentTeam(team,npcID)
	saved[team] = {}
	for i=1,3 do
		local c = frame.current[i]
		tinsert(saved[team],{c.petID,c.ability[1].abilityID,c.ability[2].abilityID,c.ability[3].abilityID})
	end
	if npcID then
		tinsert(saved[team],npcID)
	end
	selectedTeam = team
	rematch:UpdateSavedPets()
	for i=1,3 do
		frame.saved[i].shine:SetCooldown(0,1) -- do shiny cooldown ending bit
	end
end

-- check buttons keepFirst, autoShow, autoLoad
function rematch:CheckButtonOnClick()
	settings[self.setting] = self:GetChecked()
	if rematch.ReflectSetting[self.setting] then
		rematch.ReflectSetting[self.setting]()
	end
end

function rematch:CheckButtonOnEnter()
	frame.options.tooltip.text:SetText(self.tooltip)
end

-- only call from the toggle drawer button
function rematch:ToggleDrawer()
	rematch.showDrawer = not rematch.showDrawer
	settings.XPos,settings.YPos = frame:GetLeft(),frame:GetBottom()
	frame.drawer:SetShown(rematch.showDrawer)
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMLEFT",settings.XPos,settings.YPos)
end

function rematch:StartMoving()
	if not settings.LockWindow and not self.petID and not self.abilityID then
		frame:StartMoving()
	end
end

function rematch:StopMoving()
	if not settings.LockWindow and not self.petID and not self.abilityID then
		frame:StopMovingOrSizing()
		settings.XPos,settings.YPos = frame:GetLeft(),frame:GetBottom()
	end
end

--[[ Team List ]]

function rematch:TeamFormattedProperly(team)
	if type(team)~="table" then
		return
	end
	for i=1,3 do
		if not team[i] or type(team[i])~="table" then
			return
		end
	end
	return true
end

function rematch:OrderTeamList()
	wipe(rematch.teamList)
	for k,v in pairs(saved) do
		if k:match("^~temp~") or not rematch:TeamFormattedProperly(v) then
			saved[k] = nil -- remove any temporary teams
		else
			tinsert(rematch.teamList,k)
		end
	end
	table.sort(rematch.teamList,function(e1,e2) return e1:lower()<e2:lower() end)
end

function rematch:UpdateTeamList()
	local numData = #rematch.teamList
	local scrollFrame = frame.drawer.teams.scrollFrame
	local offset = HybridScrollFrame_GetOffset(scrollFrame)
	local buttons = scrollFrame.buttons
	rematch.selectedTeamVisible = nil
	for i=1, #buttons do
		local index = i + offset
		local button = buttons[i]
		button:Hide()
		button.selectedTexture:Hide()
		if ( index <= numData) then
			local team = rematch.teamList[index]
			button.team = team
			if team==selectedTeam then
				rematch.selectedTeamVisible = true
				button.selectedTexture:Show()
			end
			if saved[team][4] then
				button.name:SetTextColor(1,1,1)
			else
				button.name:SetTextColor(1,.82,0)
			end
			button.name:SetText(team)
			for j=1,3 do
				local pet = saved[team][j][1]
				local icon
				if type(pet)=="string" then -- this is an owned pet
					icon = select(9,C_PetJournal.GetPetInfoByPetID(pet))
				elseif type(pet)=="number" then -- this is a missing pet (species ID)
					icon = select(2,C_PetJournal.GetPetInfoBySpeciesID(pet))
				end
				button[tostring(j)]:SetTexture(icon or "Interface\\Buttons\\UI-PageButton-Background")
			end
			button:Show()
		end
	end
	HybridScrollFrame_Update(scrollFrame, 18*numData, 18)
	rematch:VerifyButtons()
end

--[[ Current pets ]]

-- updates the 3 current pets and their abilities
function rematch:UpdateCurrentPets()
	local petID
	rematch:WipePetFrames(frame.current)
	wipe(rematch.info)
	local ability = rematch.info
	for i=1,3 do
		petID,ability[1],ability[2],ability[3] = C_PetJournal.GetPetLoadOutInfo(i)
		if petID then
			frame.current[i].petID = petID
			local speciesID,_,level,_,_,_,_,_,icon = C_PetJournal.GetPetInfoByPetID(petID)
			-- update this pet's icon
		  frame.current[i].icon:SetTexture(icon)
			frame.current[i].icon:Show()
			for j=1,3 do
				frame.current[i].ability[j].abilityID = ability[j]
				frame.current[i].ability[j].icon:SetTexture((select(2,C_PetJournal.GetPetAbilityInfo(ability[j]))))
				frame.current[i].ability[j].icon:Show()
			end
			-- update hp and whether pet is dead
			local hp,hpmax = C_PetJournal.GetPetStats(frame.current[i].petID)
			if hp==0 then
				frame.current[i].dead:Show() -- dead pet :(
				frame.current[i].health:Hide()
			else
				frame.current[i].dead:Hide()
				frame.current[i].health:SetWidth(hp>0 and 38*(hp/hpmax) or 1)
				frame.current[i].health:Show()
			end
		end
	end
end

-- there's no event for when loadout pets change (really!) so we hooksecurefunc both
-- SetAbility and SetPetLoadOutInfo to do an update 0.1 seconds after they happen.
-- the wait is to allow multiple simultaneous changes to happen before doing an update
function rematch:WatchForChangingPets()
	if not rematch.loadoutsHooked then
		rematch.loadoutsHooked = 1
		hooksecurefunc(C_PetJournal,"SetAbility",rematch.StartPetsChanging)
		hooksecurefunc(C_PetJournal,"SetPetLoadOutInfo",rematch.StartPetsChanging)
	end
end
function rematch:StartPetsChanging()
	rematch:StartTimer("PetsChanging",0.1,rematch.UpdateCurrentPets)
end

--[[ Saving teams ]]

function rematch:TeamOnClick()
	local team = self.team
	if selectedTeam==team then
		selectedTeam = nil
	else
		selectedTeam = team
	end
	rematch:UpdateTeamList()
	rematch:UpdateSavedPets()
end

function rematch:TeamOnDoubleClick()
	selectedTeam = self.team
	rematch:UpdateTeamList()
	rematch:UpdateSavedPets()
	rematch:LoadPets(self.team)
end

-- wipes frames of any leftover pets/abilities
-- parent is likely frame.current, frame.saved or inc.pets
function rematch:WipePetFrames(parent)
	for i=1,3 do
		parent[i].petID = nil
		parent[i].icon:Hide()
		for j=1,3 do
			parent[i].ability[j].abilityID = nil
			parent[i].ability[j].icon:Hide()
		end
	end
end

function rematch:UpdateSavedPets()
	rematch:WipePetFrames(frame.saved)
	for i=1,6 do
		frame.hints[tostring(i)]:Hide()
	end
	if selectedTeam then -- then see if a team is selected
		rematch:FillSavedFromTeam(selectedTeam)
		local actualTeam = selectedTeam:gsub("~temp~","") -- remove temporary tag if it exists
		if saved[selectedTeam][4] then
			frame.savedHeader.text:SetText(format("Saved for \124cffffffff%s",actualTeam))
		else
			frame.savedHeader.text:SetText(format("Saved as \"%s\"",actualTeam))
		end
	elseif UnitExists("target") then
		if targetedNpcID then -- this is an npc
			if saved[targetedName] and saved[targetedName][4]==targetedNpc then
				rematch:FillSavedFromTeam(targetedName)
			end
			frame.savedHeader.text:SetText(format("Saved for \124cffffffff%s",targetedName))
		else -- this is a player, treat them just like having no target
			frame.savedHeader.text:SetText("Saved Battle Pets")
		end
	else -- no target, no saved team
		frame.savedHeader.text:SetText("Saved Battle Pets")
	end
	rematch:VerifyButtons()
end

-- returns name of the given unit and npcID if it's an npc, or nil if unit doesn't exist
function rematch:GetUnitNameandID(unit)
	if UnitExists(unit) then
		local name = UnitName(unit)
		local npcID = tonumber(UnitGUID(unit):sub(6,10),16)
		if npcID~=0 then
			return name,npcID -- this is an npc, return its name and npcID
		else
			return name -- this is a player, return its name
		end
	end
end

function rematch:DesaturateIfMissing(texture,missing)
	texture:SetDesaturated(missing)
	if missing then
		texture:SetVertexColor(.75,.75,.75)
	else
		texture:SetVertexColor(1,1,1)
	end
end

function rematch:FillSavedFromTeam(team)

	local attacks = rematch.attackTypes
	for i=1,10 do
		attacks[i]=0 -- reset counter of attack types
	end

	-- this part updates the displayed saved pets and their abilities
	for i=1,3 do
		local icon,speciesName
		local petID = saved[team][i][1]
		local missing
		if type(petID)=="string" then -- petID of owned pet
			icon = select(9,C_PetJournal.GetPetInfoByPetID(petID))
		elseif type(petID)=="number" then
			speciesName,icon = C_PetJournal.GetPetInfoBySpeciesID(petID)
			if speciesName then
				local _,foundPetID = C_PetJournal.FindPetIDByName(speciesName)
				if foundPetID then
					petID = foundPetID
					saved[team][i][1] = petID
				else
					missing = 1
				end
			end
		end
		if icon then
			frame.saved[i].petID = petID
			frame.saved[i].icon:SetTexture(icon)
			rematch:DesaturateIfMissing(frame.saved[i].icon,missing)
			frame.saved[i].icon:Show()
			for j=1,3 do
				local _,_,abilityIcon,_,_,_,abilityType,noHints = C_PetBattles.GetAbilityInfoByID(saved[team][i][1+j])
				if abilityIcon then
					if not noHints and abilityType then -- if ability is an attack, increment its counter
						attacks[abilityType] = attacks[abilityType]+1
					end
					frame.saved[i].ability[j].abilityID = saved[team][i][1+j]
					frame.saved[i].ability[j].icon:SetTexture(abilityIcon)
					rematch:DesaturateIfMissing(frame.saved[i].ability[j].icon,missing)
					frame.saved[i].ability[j].icon:Show()
				end
			end
		end
	end

	-- the following updates the hints window to show what the team's attacks
	-- are strong against.  we'll go through the attackTypes table gathered
	-- above and multiply the values by 100, add their index, and reverse sort.
	-- ie 4x flying is 403, 1x mechanical is 110.
	for i=1,10 do
		attacks[i] = attacks[i]*100+i
	end
	table.sort(attacks,function(e1,e2) return e1>e2 end)
	-- now the most common attacks are at the top. the value%100 will be the attack type
	for i=1,10 do
		attacks[i] = attacks[i]>99 and attacks[i]%100 or nil
	end
	frame.hints["1"]:SetShown(#attacks~=0) -- show strong/weak green arrow if there are any attacks
	local centerRow
	for i=1,5 do -- only care about first 5
		if attacks[i] then
			local strong = tostring(i+1)
			frame.hints[strong]:SetTexture("Interface\\PetBattles\\PetIcon-"..PET_TYPE_SUFFIX[rematch.attackHints[attacks[i]][1]])
			frame.hints[strong]:Show()
		elseif i==3 then
			centerRow = true
		end
	end
	-- move first row to center if less than 3 attacks to display
	for i=1,3 do
		frame.hints[tostring(i)]:SetPoint("CENTER",frame.hints,centerRow and "LEFT" or "TOPLEFT",(i-1)*20+13,centerRow and 0 or -13)
	end
end

--[[ Loading teams ]]

-- returns true if team needs loaded, nil otherwise
function rematch:PetsNeedLoading(team)
	if not team or not saved[team] then return end
	local info = rematch.info
	for i=1,3 do
		wipe(info)
		info[1],info[2],info[3],info[4] = C_PetJournal.GetPetLoadOutInfo(i)
		for j=1,4 do
			if saved[team][i][j] ~= info[j] then
				return true -- something is different, pets need loading
			end
		end
	end
end

-- this takes a 4-entry table {pet,ability1,ability2,ability3} and returns true if
-- the pet is a valid petID or speciesID and all abilities are known by the pet.
-- this only *needs* to run when loading pets. completely optional rest of the time
function rematch:ValidatePet(petTable)
	if type(petTable)~="table" or #petTable~=4 then
		return false -- not a valid pet table, test failed
	end
	local petID = petTable[1]
	local speciesID
	if type(petID)=="string" then
		speciesID = C_PetJournal.GetPetInfoByPetID(petID)
	elseif type(petID)=="number" then
		if C_PetJournal.GetPetInfoBySpeciesID(petID) then
			speciesID = petID
		end
	end
	if not speciesID then
		return -- not a valid petID or speciesID, test failed
	end
	wipe(rematch.abilityList)
	C_PetJournal.GetPetAbilityList(speciesID,rematch.abilityList,rematch.levelList)
	for i=2,4 do
		local abilityKnown = nil
		for j=1,#rematch.abilityList do
			if petTable[i]==rematch.abilityList[j] then
				abilityKnown = 1
				break
			end
		end
		if not abilityKnown then
			return -- pet has an ability it doesn't know, test failed
		end
	end
	return true -- this pet is a valid pet with all valid abilities, it passed!
end

-- loads the named team; doesn't check if it needs loaded (which is done by rematch:PetsNeedLoading(team))
-- due to pets maybe missing (speciesID or caged pet), and for KeepFirst option, it generates a table
-- (rematch.petsToLoad) and populates it what what the pets to actually load.  Then it turns on
-- the LoadProcessor which handles the actual loading.
function rematch:LoadPets(team)
	if not team or not saved[team] then return end

	rematch.preLoadCompanion = C_PetJournal.GetSummonedPetGUID() -- note current pet

	local toLoad = rematch.petsToLoad
	for i=1,3 do
		wipe(toLoad[i])
	end
	local loadIndex = 1
	-- handle KeepFirst
	if settings.KeepFirst then
		toLoad[1][1],toLoad[1][2],toLoad[1][3],toLoad[1][4] = C_PetJournal.GetPetLoadOutInfo(1)
		loadIndex = loadIndex + 1
	end
	local savedIndex = 1 -- start on first saved pet index
	while loadIndex<=3 and savedIndex<=3 do
		local skipPet -- if this pet is already assigned to an earlier slot, we'll skip it
		for i=1,loadIndex-1 do
			if saved[team][savedIndex][1] == toLoad[i][1] then
				skipPet = true
			end
		end
		-- do a more rigorous test to see if this pet is valid and has abilities that can load
		if not rematch:ValidatePet(saved[team][savedIndex]) then
			skipPet = true -- bad pet! *rolls up newspaper*
		end
		if not skipPet then
			local petID = saved[team][savedIndex][1]
			-- there's a chance this pet no longer exists (or is a speciesID)
			if type(petID)=="string" and C_PetJournal.GetPetStats(petID) then
				for i=1,4 do
					toLoad[loadIndex][i] = saved[team][savedIndex][i]
				end
				loadIndex = loadIndex + 1
			end
		end
		savedIndex = savedIndex + 1
	end

	-- at this point, petsToLoad is populated with what we want to load

	rematch.loadProcessor = rematch.loadProcessor or CreateFrame("Frame")
	rematch.loadProcessor:SetScript("OnUpdate",rematch.ProcessLoad)
end

-- this is the meat of the load process. every frame, it will go through the current
-- loadout pets and see if any pets or abilities need swapped in. if so, it will load
-- the pet or ability and immediately leave. once no swaps happen, this stops
function rematch:ProcessLoad(elapsed)
	local info = rematch.info
	local toLoad = rematch.petsToLoad
	for i=1,3 do
		wipe(info)
		info[1],info[2],info[3],info[4] = C_PetJournal.GetPetLoadOutInfo(i)
		if toLoad[i][1] and info[1]~=toLoad[i][1] then
			C_PetJournal.SetPetLoadOutInfo(i,toLoad[i][1])
			return
		elseif toLoad[i][1] then
			for j=2,4 do
				if info[j]~=toLoad[i][j] then
					C_PetJournal.SetAbility(i,j-1,toLoad[i][j])
					return
				end
			end
		end
	end
	-- if we made it this far, pets are all loaded!
	if frame:IsVisible() then
		for i=1,3 do
			frame.current[i].shine:SetCooldown(0,1) -- shine the current pets
		end
	end
	if PetJournal and PetJournal:IsVisible() then
		PetJournal_UpdatePetLoadOut() -- update petjournal if it's open
	end
	self:SetScript("OnUpdate",nil)

	if settings.KeepSummoned then
		rematch:StartTimer("RestoreCompanion",2,rematch.RestoreCompanion)
	end

	if settings.AutoShow and selectedTeam and targetedName==selectedTeam then
		frame:Hide()
	end

	rematch:ClearSelected()
end

function rematch:RestoreCompanion()
	if not InCombatLockdown() then -- can't SummonPetByGUID in combat :(
		local nowSummoned = C_PetJournal.GetSummonedPetGUID()
		if not rematch.preLoadCompanion and nowSummoned then
			C_PetJournal.SummonPetByGUID(nowSummoned) -- something summoned, had nothing before
		elseif nowSummoned ~= rematch.preLoadCompanion then
			C_PetJournal.SummonPetByGUID(rematch.preLoadCompanion) -- something summoned different than before
		end
	end
end

--[[ Timer Management ]]

rematch.timerFuncs = {} -- indexed by arbitrary name, the func to run when timer runs out
rematch.timerTimes = {} -- indexed by arbitrary name, the duration to run the timer
rematch.timerFrame = CreateFrame("Frame")
rematch.timerFrame:Hide()

function rematch:StartTimer(name,duration,func)
	rematch.timerFuncs[name] = func
	rematch.timerTimes[name] = duration
	rematch.timerFrame:Show()
end

function rematch:StopTimer(name)
	if name and rematch.timerTimes[name] then
		rematch.timerTimes[name] = nil
	end
end

rematch.timerFrame:SetScript("OnUpdate",function(self,elapsed)
	local tick
	local times = rematch.timerTimes
  for name in pairs(times) do
		times[name] = times[name] - elapsed
		if times[name] < 0 then
			times[name] = nil
			rematch.timerFuncs[name]()
		end
		tick = 1
	end
	if not tick then
		self:Hide()
	end
end)

--[[ Option relections ]]

rematch.ReflectSetting = {}

-- during initialization and when options are clicked, these are run.
-- note the checkbutton should be named the same as the setting

function rematch.ReflectSetting.LockWindow()
	if settings.LockWindow then
		frame:SetBackdropBorderColor(.5,.41,0)
		frame.separator:SetVertexColor(.5,.5,.5)
		frame.share.separator:SetVertexColor(.5,.5,.5)
		frame.options.checks.LockWindow:SetChecked(1)
	else
		frame:SetBackdropBorderColor(1,.82,0)
		frame.separator:SetVertexColor(1,1,1)
		frame.share.separator:SetVertexColor(1,1,1)
		frame.options.checks.LockWindow:SetChecked(nil)
	end
	rematch:SetESCable("RematchFrame",false) -- turn it off if it's on
end

function rematch.ReflectSetting.KeepFirst()
	frame.current[1].lock:SetChecked(settings.KeepFirst)
	frame.options.checks.KeepFirst:SetChecked(settings.KeepFirst)
	if settings.KeepFirst then
		frame.current[1].lock:Show()
	elseif not MouseIsOver(frame.current[1]) then
		frame.current[1].lock:Hide()
	end
end

function rematch.ReflectSetting.LargeWindow()
	-- instead of a straight SetScale, we'll adjust position so it stays at the same
	-- relative position on the screen, growing out/in from the bottomleft corner
	local oldScale = frame:GetEffectiveScale()
	settings.XPos = frame:GetLeft()
	settings.YPos = frame:GetBottom()
	frame:SetScale(settings.LargeWindow and 1.25 or 1.0)
	local newScale = frame:GetEffectiveScale()
	settings.XPos = (settings.XPos*oldScale)/newScale
	settings.YPos = (settings.YPos*oldScale)/newScale
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMLEFT",settings.XPos,settings.YPos)
	frame.options.checks.LargeWindow:SetChecked(settings.LargeWindow)
	RematchFloatingPetCard:SetScale(settings.LargeWindow and 1.2 or 1)
end

function rematch.ReflectSetting.KeepSummoned()
	frame.options.checks.KeepSummoned:SetChecked(settings.KeepSummoned)
end

function rematch.ReflectSetting.AutoShow()
	frame.options.checks.AutoShow:SetChecked(settings.AutoShow)
end

function rematch.ReflectSetting.AutoLoad()
	frame.options.checks.AutoLoad:SetChecked(settings.AutoLoad)
end

function rematch.ReflectSetting.DisableShare()
	frame.options.checks.DisableShare:SetChecked(settings.DisableShare)
	if settings.DisableShare then
		frame.drawer.sendButton:Disable()
	else
		frame.drawer.sendButton:Enable()
	end
end

function rematch:ClearSelected()
	selectedTeam = nil
	targetedName = nil
	targetedNpcID = nil
	rematch:UpdateSavedPets()
	if frame.drawer:IsVisible() then
		rematch:UpdateTeamList()
	end
end

--[[ Pet Ability Tooltips ]]

-- ability tooltips use FloatingPetBattleAbilityTooltip with a few tweaks
function rematch:DisplayAbilityTooltip()
	rematch.borderKeys = rematch.borderKeys or {"BorderTopLeft","BorderTopRight","BorderBottomLeft","BorderBottomRight","BorderLeft","BorderRight","BorderTop","BorderBottom"}
	local petID = self:GetParent().petID
	if petID and self.abilityID then
		local _,maxHealth,power,speed
		if type(petID)=="string" then
			_,maxHealth,power,speed = C_PetJournal.GetPetStats(petID)
		else
			maxHealth,power,speed = 100,0,0 -- missing pets are weak!
		end
		FloatingPetBattleAbility_Show(self.abilityID,maxHealth,power,speed)
		FloatingPetBattleAbilityTooltip:ClearAllPoints()
		FloatingPetBattleAbilityTooltip:SetPoint("BOTTOMLEFT",self,"BOTTOMRIGHT")
		-- make border gold (Blizzard why you no use tooltip Backdrop)
		for _,key in pairs(rematch.borderKeys) do
			FloatingPetBattleAbilityTooltip[key]:SetVertexColor(.9,.738,0)
		end
		-- hide close button that can't be clicked anyway
		FloatingPetBattleAbilityTooltip.CloseButton:Hide()
		-- if LargeWindow not checked, shrink tooltip by 10%
		FloatingPetBattleAbilityTooltip:SetScale(settings.LargeWindow and 1 or .9)
	end
end
function rematch:HideAbilityTooltip()
	FloatingPetBattleAbilityTooltip:Hide()
	-- revert gold border, show close button again and restore scale
	for _,key in pairs(rematch.borderKeys) do
		FloatingPetBattleAbilityTooltip[key]:SetVertexColor(1,1,1)
	end
	FloatingPetBattleAbilityTooltip.CloseButton:Show()
	FloatingPetBattleAbilityTooltip:SetScale(1)
end

--[[ Sharing ]]

-- converts a team into a string and assigns it to rematch.shareString
-- used for export and send
function rematch:GenerateShareString(team)
	if team:len()>128 then
		team = strsub(1,128)
	end
	local info = rematch.info
	wipe(info)
	tinsert(info,team)
	tinsert(info,saved[team][4] or 0)
	for i=1,3 do
		local speciesID
		if type(saved[team][i][1])=="string" then
			speciesID = C_PetJournal.GetPetInfoByPetID(saved[team][i][1])
		else
			speciesID = saved[team][i][1]
		end
		tinsert(info,speciesID)
		for j=2,4 do
			tinsert(info,saved[team][i][j])
		end
	end
	rematch.shareString = format("%s:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:",unpack(info))
end

-- parses a shareString (import string or received via CHAT_MSG_ADDON)
-- if no teamTable is given, returns the name of the team within the string, or nil if a bad string
-- otherwise converts the string into a team-formatted table ["Name"]={{p1,a1,a2,a3},{p2,a1,a2,a3},{p3,a1,a2,a3},npc}
-- NOTE: that teamTable is the PARENT of the key being created/overwritten, not the team itself
-- since the key is generated from the string.
-- returns the name of the team/key if string loaded ok
function rematch:ParseShareString(shareString,teamTable)

	local name,npcID,theRest = shareString:match("^(.-):(%d+):(%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+:)")

	if name and teamTable then -- skipping this if not a good string or only wanted name

		if teamTable[name] then -- team already existed, wipe it
			for i=1,3 do
				wipe(teamTable[name][i])
			end
			teamTable[name][4] = nil
		else -- new team, create new table entry for this name
			teamTable[name] = {{},{},{}}
		end

		local s=teamTable[name]
		s[1][1],s[1][2],s[1][3],s[1][4],s[2][1],s[2][2],s[2][3],s[2][4],s[3][1],s[3][2],s[3][3],s[3][4] = strsplit(":",theRest)
		for i=1,3 do
			for j=1,4 do
				s[i][j] = tonumber(s[i][j])
			end
			if not rematch:ValidatePet(s[i]) then -- if this is not a valid pet
				for j=1,4 do
					s[i][j] = 0 -- zero its entries (a 0 is the cross section of a rolled up newspaper)
				end
			end
		end
		if npcID~="0" then -- add npc id if it exists
			tinsert(s,tonumber(npcID))
		end

	end

	return name
end

function rematch:ExportTeam()
	rematch.shareMode = 1 -- export
	rematch:GenerateShareString(selectedTeam)
	rematch:DetailBoxOnChange() -- set and highlight export string
	frame.share:Show()
end

function rematch:ImportTeam()
	rematch.shareMode=2
	frame.share:Show()
	rematch:ClearSelected() -- we'll fill the saved/selected team with what we may import
end

function rematch:DetailBoxOnChange()
	if rematch.shareMode==1 then
		-- in export mode we reset and re-highlight the string on any changes
		local export = frame.share.detailBox.text
		export:SetText(rematch.shareString)
		export:HighlightText()
		export:SetCursorPosition(0)
		return
	end
	-- if this is an import or send, clear any warnings when box emptied
	local detail = frame.share.detailBox.text:GetText()
	detail = (detail or ""):trim()
	if detail:len()==0 then
		rematch:SetShareStatus()
		frame.share.saveButton:Disable()
		if rematch.shareMode==2 and selectedTeam then
			rematch:ClearSelected() -- also clear any team generated, if any
		end
		return
	end
	if rematch.shareMode==2 then
		-- in import mode we verify that this text is in an import-able format
		-- if so, generate a temporary team and select it
		local name = rematch:ParseShareString(detail)
		if not name then -- verify this is a valid team format
			rematch:SetShareStatus("Sorry, this is not a valid team.")
			frame.share.saveButton:Disable()
			if selectedTeam then
				rematch:ClearSelected()
			end
			return
		elseif saved[name] then
			rematch:SetShareStatus("A team already has this name. Saving will overwrite it.")
		else
			rematch:SetShareStatus()
		end
		frame.share.saveButton:Enable()
		-- create a temporary team and display as saved pets
		-- this creates a little garbage if they manually type stuff in, but not going to worry; this is for pasting
		local tempTeamName = "~temp~"..detail -- prefixing import string with "~temp~" to make a temporary team
		selectedTeam = rematch:ParseShareString(tempTeamName,saved) -- populate temp team from string
		rematch:UpdateSavedPets(1)

	elseif rematch.shareMode==3 then -- send mode
		-- if a status is up, clear it when any changes made to entered name
		if frame.share.statusIcon:IsVisible() then
			rematch:SetShareStatus()
		end
		frame.share.saveButton:Enable() -- if we're here, detail string length is > 0
	end
end

-- displays a message and an icon in the lowerleft corner of frame.share and stops waiting for stuff
-- if no message is given, the status is hidden
-- if no icon is given, then the warning icon is used
function rematch:SetShareStatus(message,icon)
	frame.share.statusIcon:SetTexture(icon and icon or "Interface\\DialogFrame\\UI-Dialog-Icon-AlertNew")
	frame.share.statusText:SetText(message or "")
	frame.share.statusIcon:SetShown(message and true or false)
	frame.share.statusText:SetShown(message and true or false)
	frame:UnregisterEvent("CHAT_MSG_SYSTEM")
	rematch:StopTimer("SendTimeout")
end

-- when enter pressed while share.detailBox.text has focus, or save/send button hit
function rematch:SaveShare()
	if rematch.shareMode==1 then -- exporting
		frame.share:Hide()
	elseif rematch.shareMode==2 then -- importing
		if frame.share.saveButton:IsEnabled() then
			local actualTeam = selectedTeam:gsub("~temp~","")
			saved[actualTeam] = {{},{},{}}
			for i=1,3 do
				for j=1,4 do
					saved[actualTeam][i][j] = saved[selectedTeam][i][j]
				end
			end
			saved[actualTeam][4] = saved[selectedTeam][4]
			selectedTeam = actualTeam
			frame.share:Hide()
			for i=1,3 do
				frame.saved[i].shine:SetCooldown(0,1) -- do shiny cooldown ending bit
			end
		end
	elseif rematch.shareMode==3 then -- sending
		rematch:GenerateShareString(selectedTeam)
		if rematch.shareString then
			rematch:SetShareStatus("Sending...","Interface\\Icons\\SOR-mail")
			rematch.sendee = frame.share.detailBox.text:GetText():trim()
			frame:RegisterEvent("CHAT_MSG_SYSTEM")
			rematch:StartTimer("SendTimeout",3,rematch.SendTimeout)
			SendAddonMessage("Rematch",rematch.shareString,"WHISPER",rematch.sendee)
		end
	end
end

-- this is registered when we send a team, and unregistered on next SetShareStatus (including panel closing)
function rematch:CHAT_MSG_SYSTEM(message)
	if message==format(ERR_CHAT_PLAYER_NOT_FOUND_S,rematch.sendee) then
		rematch:SetShareStatus("They do not appear to be online.")
	end
end

-- this is run on a 3-second delay after a send, unless a status update shuts it down before then
function rematch:SendTimeout()
	rematch:SetShareStatus("No response. Due to lag or no Rematch?")
end

-- SendAddonMessage from other users to us will trigger this event
function rematch:CHAT_MSG_ADDON(prefix,message,_,sender)
	if prefix~="Rematch" then
		return
	end
	if message=="ok" then -- sendee received team and sent back an ok
		rematch:SetShareStatus("The team was received!","Interface\\RaidFrame\\ReadyCheck-Ready")
	elseif message=="busy" then -- sendee received team but is already looking at a team
		rematch:SetShareStatus("They have another team waiting. Try again later.")
	elseif message=="combat" then -- sendee received team, but they're in combat
		rematch:SetShareStatus("They're in combat. Try again later.")
	elseif message=="block" then -- sendee received team, but has settings.DisableShare checked
		rematch:SetShareStatus("They have team sharing disabled.")
	else -- it's an incoming team

		-- if our DisableShare is set, tell the sender and leave
		if settings.DisableShare then
			SendAddonMessage("Rematch","block","WHISPER",sender)
			return
		end
		-- if user is already looking at a team, tell the sender and leave
		if RematchIncomingFrame:IsVisible() then
			SendAddonMessage("Rematch","busy","WHISPER",sender)
			return
		end
		-- if user is in combat (not a game limitation, but it'd be very annoying to get incoming pets in a boss fight)
		if InCombatLockdown() then
			SendAddonMessage("Rematch","combat","WHISPER",sender)
			return
		end
		-- otherwise send back "ok"
		SendAddonMessage("Rematch","ok","WHISPER",sender)
		-- now display the team
		local inc = RematchIncomingFrame

		inc.greeting:SetText("\124cffffd200"..sender.."\124cffffffff has sent you a battle pet team!")

		if not inc.pets then -- if we haven't received a pet this session, set up some stuff
			inc.header.text:SetText("Incoming Rematch Team")
			inc.warning:SetText("A team with this name already exists.\nEdit the name or hit Save to overwrite it.")
			inc.pets = {}
			for i=1,3 do
				inc.pets[i] = CreateFrame("Button",nil,inc,"RematchIncomingPetTemplate")
				inc.pets[i]:SetPoint("TOPLEFT",(i-1)*78+16,-70)
				inc.pets[i].ability = {}
				for j=1,3 do
					inc.pets[i].ability[j] = CreateFrame("Button",nil,inc.pets[i],"RematchSmallSlotTemplate")
					inc.pets[i].ability[j]:SetPoint("TOPLEFT",inc.pets[i],"TOPRIGHT",2,(j-1)*-22+10)
				end
			end
		end

		-- parse incoming team into rematch.incomingTeam
		wipe(rematch.incomingTeam)
		rematch.incomingName = rematch:ParseShareString(message,rematch.incomingTeam)

		if not rematch.incomingName then
			return -- we weren't sent a valid team, don't tell sender. they could be using a future version
		end

		inc.incomingAs:SetText(rematch.incomingName) -- put name in editbox
		inc.incomingAs:SetCursorPosition(0)

		rematch:WipePetFrames(inc.pets)

		-- fill slots with pets from generated team
		local team = rematch.incomingTeam[rematch.incomingName]
		for i=1,3 do
			local speciesName,icon = C_PetJournal.GetPetInfoBySpeciesID(team[i][1])
			if speciesName then -- all incoming petIDs are speciesIDs
				local _,foundPetID = C_PetJournal.FindPetIDByName(speciesName)
				if foundPetID then
					team[i][1] = foundPetID
				end
				inc.pets[i].petID = team[i][1]
				inc.pets[i].icon:SetTexture(icon)
				inc.pets[i].icon:Show()
				rematch:DesaturateIfMissing(inc.pets[i].icon,not foundPetID)
				for j=1,3 do
					inc.pets[i].ability[j].abilityID = team[i][j+1]
					inc.pets[i].ability[j].icon:SetTexture((select(3,C_PetBattles.GetAbilityInfoByID(team[i][j+1]))))
					inc.pets[i].ability[j].icon:Show()
					rematch:DesaturateIfMissing(inc.pets[i].ability[j].icon,not foundPetID)
				end
			end
		end
		inc:Show()
		rematch.IncomingAsOnTextChanged(inc.incomingAs)
	end
end

-- when incoming editbox changes
function rematch:IncomingAsOnTextChanged()
	local inc = RematchIncomingFrame
	local name = (self:GetText() or ""):trim()
	inc.allClear:Show()
	inc.warning:Hide()
	if name:len()==0 then
		inc.saveButton:Disable()
		return
	elseif saved[name] then
		inc.allClear:Hide()
		inc.warning:Show()
	end
	inc.saveButton:Enable()
end

-- final save from incoming dialog; create a saved[] team and update drawer if it's up
function rematch:SaveIncoming()
	if RematchIncomingFrame.saveButton:IsEnabled() then
		-- name the team what's currently in the incomingAs editbox
		local name = RematchIncomingFrame.incomingAs:GetText():trim()
		saved[name] = {{},{},{}}
		local team = saved[name]
		-- copy the incomingTeam to the live saved team
		for i=1,3 do
			for j=1,4 do
				team[i][j] = rematch.incomingTeam[rematch.incomingName][i][j]
			end
		end
		if rematch.incomingNpc~=0 then
			team[4] = rematch.incomingNpc
		end
		-- select it, update team list if needed and dismiss incoming frame
		selectedTeam = name
		if frame.drawer:IsVisible() then
			rematch:OrderTeamList()
			rematch:UpdateTeamList()
		end
		RematchIncomingFrame:Hide()
	end
end

SlashCmdList["REMATCH"] = frame.Toggle
SLASH_REMATCH1 = "/rematch"